Imports System.Xml
Imports System.ComponentModel

''' <summary>
''' Test collection (IList) object that contains
''' OfficeItem objects.
''' </summary>
Public Class OfficeItems
  Inherits CollectionBase

  Implements IBindingList

  ''' <summary>
  ''' Returns the item at the specified index.
  ''' </summary>
  ''' <param name="index">Index of the item to return.</param>
  ''' <returns>The OfficeItem object at the specified index.</returns>
  Default Public ReadOnly Property Item(ByVal index As Integer) As OfficeItem
    Get
      Return DirectCast(list(index), OfficeItem)
    End Get
  End Property

  ''' <summary>
  ''' Load the OfficeItem child objects from an XML file.
  ''' </summary>
  Public Sub LoadData()

    list.Clear()
    Dim reader As New XmlTextReader("..\data.xml")
    Try
      With reader
        While .Read
          If .NodeType = XmlNodeType.Element Then
            Select Case .Name
              Case "item"
                Dim name As String = .GetAttribute("name")
                Dim type As String = .GetAttribute("type")
                Dim item As New OfficeItem(Me, name, type)
                list.Add(item)
                AddHandler item.PropertyChanged, AddressOf ChildChanged
            End Select
          End If
        End While
      End With
    Finally
      reader.Close()
    End Try

    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.Reset, 0))

  End Sub

  ''' <summary>
  ''' Handles PropertyChanged events for all child objects.
  ''' </summary>
  ''' <param name="sender">The changed child object.</param>
  ''' <param name="e">The property that was changed.</param>
  Private Sub ChildChanged(ByVal sender As Object, ByVal e As PropertyChangedEventArgs)

    ' find the item's index
    Dim index As Integer
    For Each item As Object In list
      If item.Equals(sender) Then
        Exit For
      End If
      index += 1
    Next

    ' indicate item has changed
    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.ItemChanged, index))

  End Sub

#Region " List event handlers "

  ''' <summary>
  ''' Makes sure event handlers are disconnected
  ''' before items are removed.
  ''' </summary>
  Protected Overrides Sub OnClear()

    For Each item As OfficeItem In list
      RemoveHandler item.PropertyChanged, AddressOf ChildChanged
    Next

  End Sub

  ''' <summary>
  ''' Raises event to indicate list has changed.
  ''' </summary>
  Protected Overrides Sub OnClearComplete()
    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.Reset, 0))
  End Sub

  ''' <summary>
  ''' Adds event handler for newly added item. Raises event to
  ''' indicate list has changed.
  ''' </summary>
  Protected Overrides Sub OnInsertComplete(ByVal index As Integer, ByVal value As Object)
    AddHandler DirectCast(value, OfficeItem).PropertyChanged, AddressOf ChildChanged
    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.ItemAdded, index))
  End Sub

  ''' <summary>
  ''' Removes event handler for removed object. Raises event to
  ''' indicate list has changed.
  ''' </summary>
  Protected Overrides Sub OnRemoveComplete(ByVal index As Integer, ByVal value As Object)
    RemoveHandler DirectCast(value, OfficeItem).PropertyChanged, AddressOf ChildChanged
    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.ItemDeleted, index))
  End Sub

  ''' <summary>
  ''' Removes event handler for removed object, and adds event
  ''' handler for newly added item. Raises event to indicate
  ''' list has changed.
  ''' </summary>
  Protected Overrides Sub OnSetComplete(ByVal index As Integer, ByVal oldValue As Object, ByVal newValue As Object)
    RemoveHandler DirectCast(oldValue, OfficeItem).PropertyChanged, AddressOf ChildChanged
    AddHandler DirectCast(newValue, OfficeItem).PropertyChanged, AddressOf ChildChanged
    RaiseEvent ListChanged(Me, New ListChangedEventArgs(ListChangedType.ItemChanged, index))
  End Sub

#End Region

#Region " IBindingList "

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public Sub AddIndex(ByVal [property] As System.ComponentModel.PropertyDescriptor) Implements System.ComponentModel.IBindingList.AddIndex

  End Sub

  ''' <summary>
  ''' Adds a new item to the list.
  ''' </summary>
  ''' <remarks>
  ''' This is implemented to support the DataGrid so it can
  ''' add items to the list dynamically.
  ''' </remarks>
  Public Function AddNew() As Object Implements System.ComponentModel.IBindingList.AddNew
    Dim item As New OfficeItem(Me)
    list.Add(item)
    Return item
  End Function

  ''' <summary>
  ''' Removes a child from the list.
  ''' </summary>
  ''' <remarks>
  ''' This is implemented to support the DataGrid so it can
  ''' add items to the list dynamically. If an item is added
  ''' dynamically and then the user presses Esc the item will
  ''' request that it be removed from the list by calling this
  ''' method. Look at the CancelEdit method in OfficeItem to
  ''' see when this method is called.
  ''' </remarks>
  ''' <param name="child">The child to remove.</param>
  Friend Sub RemoveChild(ByVal child As OfficeItem)
    Dim index As Integer
    For Each item As OfficeItem In list
      If item.Equals(child) Then
        Exit For
      End If
      index += 1
    Next
    list.RemoveAt(index)
  End Sub

  ''' <summary>
  ''' Returns True so we support in-place editing.
  ''' </summary>
  Public ReadOnly Property AllowEdit() As Boolean Implements System.ComponentModel.IBindingList.AllowEdit
    Get
      Return True
    End Get
  End Property

  ''' <summary>
  ''' Returns True so we support dynamic adding of child objects.
  ''' </summary>
  Public ReadOnly Property AllowNew() As Boolean Implements System.ComponentModel.IBindingList.AllowNew
    Get
      Return True
    End Get
  End Property

  ''' <summary>
  ''' Returns True so we support dynamic removal of child objects.
  ''' </summary>
  Public ReadOnly Property AllowRemove() As Boolean Implements System.ComponentModel.IBindingList.AllowRemove
    Get
      Return True
    End Get
  End Property

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public Sub ApplySort(ByVal [property] As System.ComponentModel.PropertyDescriptor, ByVal direction As System.ComponentModel.ListSortDirection) Implements System.ComponentModel.IBindingList.ApplySort

  End Sub

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public Function Find(ByVal [property] As System.ComponentModel.PropertyDescriptor, ByVal key As Object) As Integer Implements System.ComponentModel.IBindingList.Find

  End Function

  ''' <summary>
  ''' Return False. This collection is not sortable.
  ''' </summary>
  Public ReadOnly Property IsSorted() As Boolean Implements System.ComponentModel.IBindingList.IsSorted
    Get
      Return False
    End Get
  End Property

  ''' <summary>
  ''' Raised to indicate that data in the list has changed.
  ''' </summary>
  Public Event ListChanged(ByVal sender As Object, ByVal e As System.ComponentModel.ListChangedEventArgs) Implements System.ComponentModel.IBindingList.ListChanged

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public Sub RemoveIndex(ByVal [property] As System.ComponentModel.PropertyDescriptor) Implements System.ComponentModel.IBindingList.RemoveIndex

  End Sub

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public Sub RemoveSort() Implements System.ComponentModel.IBindingList.RemoveSort

  End Sub

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public ReadOnly Property SortDirection() As System.ComponentModel.ListSortDirection Implements System.ComponentModel.IBindingList.SortDirection
    Get
      Return ListSortDirection.Ascending
    End Get
  End Property

  ''' <summary>
  ''' Not implemented.
  ''' </summary>
  Public ReadOnly Property SortProperty() As System.ComponentModel.PropertyDescriptor Implements System.ComponentModel.IBindingList.SortProperty
    Get

    End Get
  End Property

  ''' <summary>
  ''' Returns True. This object will raise the ListChanged event.
  ''' </summary>
  Public ReadOnly Property SupportsChangeNotification() As Boolean Implements System.ComponentModel.IBindingList.SupportsChangeNotification
    Get
      Return True
    End Get
  End Property

  ''' <summary>
  ''' Returns False. This object does not implement searching.
  ''' </summary>
  Public ReadOnly Property SupportsSearching() As Boolean Implements System.ComponentModel.IBindingList.SupportsSearching
    Get
      Return False
    End Get
  End Property

  ''' <summary>
  ''' Returns False. This object does not implement sorting.
  ''' </summary>
  Public ReadOnly Property SupportsSorting() As Boolean Implements System.ComponentModel.IBindingList.SupportsSorting
    Get
      Return False
    End Get
  End Property

#End Region

End Class
